import functools
import os
import shutil
import tempfile


class temporary_directory(object):

    def __init__(self, prefix=''):
        self.prefix = prefix

    def __enter__(self):
        self.original_cwd = os.getcwd()
        self.temp_path = tempfile.mkdtemp(prefix=self.prefix)
        os.chdir(self.temp_path)
        return self.temp_path

    def __exit__(self, exc_type, exc_value, traceback):
        # in Windows, current directory needs to be changed back to release the file handle.
        if self.original_cwd and os.path.exists(self.original_cwd):
            os.chdir(self.original_cwd)
        if self.temp_path and os.path.exists(self.temp_path):
            shutil.rmtree(self.temp_path)


def in_temporary_directory(f):
    @functools.wraps(f)
    def decorated(*args, **kwds):
        with temporary_directory() as directory:
            try:
                from inspect import getfullargspec as getargspec
            except ImportError:
                from inspect import getargspec
            # If it takes directory of kwargs and kwds does already have
            # directory, inject it
            if 'directory' not in kwds and 'directory' in getargspec(f)[0]:
                kwds['directory'] = directory
            return f(*args, **kwds)
    decorated.__name__ = f.__name__
    return decorated



import os.path
# Redirect stderr to stdout to suppress output in tests
import sys
import unittest

import xml.dom.minidom as dom
from xml.parsers.expat import ExpatError

from catkin_pkg.package import (
    _check_known_attributes,
    _get_package_xml,
    Dependency,
    Export,
    has_ros_schema_reference_string,
    InvalidPackage,
    License,
    Package,
    parse_package,
    parse_package_string,
    Person,
)

try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

sys.stderr = sys.stdout

test_data_dir = os.path.join(os.path.dirname(__file__), 'data', 'package')


class PackageTest(unittest.TestCase):

    if sys.version_info[0] == 2:
        assertRaisesRegex = unittest.TestCase.assertRaisesRegexp

    def get_maintainer(self):
        maint = Mock()
        maint.email = 'foo@bar.com'
        maint.name = 'John Foo'
        return maint

    def get_group_dependency(self, name):
        group = Mock()
        group.name = name
        return group

    def test_init(self):
        maint = self.get_maintainer()
        pack = Package(name='foo',
                       version='0.0.0',
                       maintainers=[maint],
                       licenses=['BSD'])
        self.assertEqual(None, pack.filename)
        self.assertEqual('0.0.0', pack.version)
        self.assertEqual(None, pack.version_compatibility)
        self.assertEqual([], pack.urls)
        self.assertEqual([], pack.authors)
        self.assertEqual([maint], pack.maintainers)
        self.assertEqual(['BSD'], pack.licenses)
        self.assertEqual([None], [license_.file for license_ in pack.licenses])
        self.assertEqual([], pack.build_depends)
        self.assertEqual([], pack.buildtool_depends)
        self.assertEqual([], pack.run_depends)
        self.assertEqual([], pack.test_depends)
        self.assertEqual([], pack.conflicts)
        self.assertEqual([], pack.replaces)
        self.assertEqual([], pack.exports)
        self.assertEqual([], pack.group_depends)
        self.assertEqual([], pack.member_of_groups)
        pack = Package('foo',
                       name='bar',
                       version='0.0.0',
                       licenses=['BSD'],
                       maintainers=[self.get_maintainer()])
        self.assertEqual('foo', pack.filename)

        self.assertRaises(TypeError, Package, unknownattribute=42)

    def test_init_dependency(self):
        dep = Dependency('foo',
                         version_lt=1,
                         version_lte=2,
                         version_eq=3,
                         version_gte=4,
                         version_gt=5,
                         condition='$foo == 23 and $bar != 42')
        self.assertEqual('foo', dep.name)
        self.assertEqual(1, dep.version_lt)
        self.assertEqual(2, dep.version_lte)
        self.assertEqual(3, dep.version_eq)
        self.assertEqual(4, dep.version_gte)
        self.assertEqual(5, dep.version_gt)
        self.assertFalse(dep.evaluate_condition({'foo': 23, 'bar': 42}))
        self.assertFalse(dep.evaluated_condition)
        self.assertTrue(dep.evaluate_condition({'foo': 23, 'bar': 43}))
        self.assertTrue(dep.evaluated_condition)
        self.assertRaises(TypeError, Dependency, 'foo', unknownattribute=42)

        d = {}
        d[dep] = None
        dep2 = Dependency('foo',
                          version_lt=1,
                          version_lte=2,
                          version_eq=3,
                          version_gte=4,
                          version_gt=5,
                          condition='$foo == 23 and $bar != 42')
        dep2.evaluate_condition({'foo': 23, 'bar': 43})
        d[dep2] = None
        self.assertEqual(len(d), 1)
        dep3 = Dependency('foo',
                          version_lt=1,
                          version_lte=2,
                          version_eq=3,
                          version_gte=4,
                          version_gt=6)
        d[dep3] = None
        self.assertEqual(len(d), 2)

        dep = Dependency('foo', condition='foo > bar and bar < baz')
        self.assertTrue(dep.evaluate_condition({}))

        dep = Dependency('foo', condition='foo <= bar or bar >= baz')
        self.assertFalse(dep.evaluate_condition({}))

        dep = Dependency('foo', condition='$foo == ""')
        self.assertTrue(dep.evaluate_condition({}))
        self.assertFalse(dep.evaluate_condition({'foo': 'foo'}))

        dep = Dependency('foo', condition='$foo == "foo \' bar"')
        self.assertTrue(dep.evaluate_condition({'foo': "foo ' bar"}))
        self.assertFalse(dep.evaluate_condition({}))

        dep = Dependency('foo', condition="$foo == ''")
        self.assertTrue(dep.evaluate_condition({}))
        self.assertFalse(dep.evaluate_condition({'foo': 'foo'}))

        dep = Dependency('foo', condition="$foo == 'foo \" bar'")
        self.assertTrue(dep.evaluate_condition({'foo': 'foo " bar'}))
        self.assertFalse(dep.evaluate_condition({}))

        # Testing for more than 1 conditions
        dep = Dependency('foo', condition='foo > bar and bar < baz and foo > bar')
        self.assertTrue(dep.evaluate_condition({}))

        dep = Dependency('foo', condition='foo <= bar and bar >= baz and foo > bar')
        self.assertFalse(dep.evaluate_condition({}))

        dep = Dependency('foo', condition='foo > bar or bar < baz or foo <= bar')
        self.assertTrue(dep.evaluate_condition({}))

        dep = Dependency('foo', condition='foo <= bar or bar >= baz or foo <= bar')
        self.assertFalse(dep.evaluate_condition({}))

        dep = Dependency('foo', condition='foo <= bar and bar < baz or foo > bar')
        self.assertTrue(dep.evaluate_condition({}))

        dep = Dependency('foo', condition='foo <= bar or bar < baz and foo < bar')
        self.assertFalse(dep.evaluate_condition({}))

        dep = Dependency('foo', condition='foo > bar and bar >= baz or foo > bar')
        self.assertTrue(dep.evaluate_condition({}))

        dep = Dependency('foo', condition='foo <= bar or bar >= baz and foo < bar')
        self.assertFalse(dep.evaluate_condition({}))

    def test_dependency_repr(self):
        dep = Dependency('foo', condition='$foo == 2')
        assert repr(dep) == "Dependency(name='foo', condition='$foo == 2')"
        self.assertTrue(dep.evaluate_condition({'foo': 2}))
        assert repr(dep) == "Dependency(name='foo', condition='$foo == 2', evaluated_condition=True)"
        self.assertFalse(dep.evaluate_condition({'foo': 3}))
        assert repr(dep) == "Dependency(name='foo', condition='$foo == 2', evaluated_condition=False)"

    def test_init_kwargs_string(self):
        pack = Package('foo',
                       name='bar',
                       package_format='1',
                       version='0.0.1',
                       version_compatibility='0.0.0',
                       description='pdesc',
                       licenses=['BSD'],
                       maintainers=[self.get_maintainer()])
        self.assertEqual('foo', pack.filename)
        self.assertEqual('bar', pack.name)
        self.assertEqual('1', pack.package_format)
        self.assertEqual('0.0.0', pack.version_compatibility)
        self.assertEqual('0.0.1', pack.version)
        self.assertEqual('pdesc', pack.description)

    def test_init_kwargs_object(self):
        mmain = [self.get_maintainer(), self.get_maintainer()]
        mlis = ['MIT', License('BSD', 'LICENSE')]
        mauth = [self.get_maintainer(), self.get_maintainer()]
        murl = [Mock(), Mock()]
        mbuilddep = [Mock(), Mock()]
        mbuildtooldep = [Mock(), Mock()]
        mrundep = [Mock(), Mock()]
        mtestdep = [Mock(), Mock()]
        mconf = [Mock(), Mock()]
        mrepl = [Mock(), Mock()]
        mexp = [Mock(), Mock()]
        mgroup = [
            self.get_group_dependency('group1'),
            self.get_group_dependency('group2')]
        mmember = ['member1', 'member2']
        pack = Package(package_format='3',
                       name='bar',
                       version='0.0.0',
                       maintainers=mmain,
                       licenses=mlis,
                       urls=murl,
                       authors=mauth,
                       build_depends=mbuilddep,
                       buildtool_depends=mbuildtooldep,
                       run_depends=mrundep,
                       test_depends=mtestdep,
                       conflicts=mconf,
                       replaces=mrepl,
                       group_depends=mgroup,
                       member_of_groups=mmember,
                       exports=mexp)
        self.assertEqual(mmain, pack.maintainers)
        self.assertEqual(mlis, pack.licenses)
        self.assertEqual([None, 'LICENSE'], [license_.file for license_ in pack.licenses])
        self.assertEqual(murl, pack.urls)
        self.assertEqual(mauth, pack.authors)
        self.assertEqual(mbuilddep, pack.build_depends)
        self.assertEqual(mbuildtooldep, pack.buildtool_depends)
        # since run_depends are getting stores as build_export_depends as well as exec_depends
        # and the dependency objects are being cloned only the double count can be checked for
        self.assertEqual(2 * len(mrundep), len(pack.run_depends))
        self.assertEqual(mtestdep, pack.test_depends)
        self.assertEqual(mconf, pack.conflicts)
        self.assertEqual(mrepl, pack.replaces)
        self.assertEqual(mexp, pack.exports)
        self.assertEqual(mgroup, pack.group_depends)
        self.assertEqual(mmember, pack.member_of_groups)

    def test_validate_package(self):
        maint = self.get_maintainer()
        pack = Package('foo',
                       name='bar_2go',
                       package_format='1',
                       version='0.0.1',
                       description='pdesc',
                       licenses=['BSD'],
                       maintainers=[maint])
        pack.validate()

        # names that should error
        pack.name = 'bar bza'
        self.assertRaises(InvalidPackage, Package.validate, pack)
        pack.name = 'foo%'
        self.assertRaises(InvalidPackage, Package.validate, pack)

        # names that should throw warnings
        pack.name = '2bar'
        warnings = []
        pack.validate(warnings=warnings)
        self.assertIn('naming conventions', warnings[0])

        pack.name = 'bar-bza'
        warnings = []
        pack.validate(warnings=warnings)
        self.assertEqual(warnings, [])

        pack.name = 'BAR'
        warnings = []
        pack.validate(warnings=warnings)
        self.assertIn('naming conventions', warnings[0])

        # dashes are permitted for a non-catkin package
        pack.exports.append(Export('build_type', 'other'))
        pack.name = 'bar-bza'
        warnings = []
        pack.validate(warnings=warnings)
        self.assertEqual(warnings, [])
        pack.exports.pop()

        # check authors emails
        pack.name = 'bar'
        auth1 = Mock()
        auth2 = Mock()
        auth2.validate.side_effect = InvalidPackage('foo')
        pack.authors = [auth1, auth2]
        self.assertRaises(InvalidPackage, Package.validate, pack)
        pack.authors = []
        pack.validate()

        # check maintainer required with email
        pack.maintainers = []
        self.assertRaises(InvalidPackage, Package.validate, pack)
        pack.maintainers = [maint]
        maint.email = None
        self.assertRaises(InvalidPackage, Package.validate, pack)
        maint.email = 'foo@bar.com'

        for dep_type in [pack.build_depends, pack.buildtool_depends, pack.build_export_depends, pack.buildtool_export_depends, pack.exec_depends, pack.test_depends, pack.doc_depends]:
            pack.validate()
            depend = Dependency(pack.name)
            dep_type.append(depend)
            self.assertRaises(InvalidPackage, Package.validate, pack)
            dep_type.remove(depend)

    def test_invalid_package_exception(self):
        try:
            raise InvalidPackage('foo')
        except InvalidPackage as e:
            self.assertEqual('foo', str(e))
            self.assertEqual(None, e.package_path)
        try:
            raise InvalidPackage('foo', package_path='./bar')
        except InvalidPackage as e:
            self.assertEqual("Error(s) in package './bar':\nfoo", str(e))
            self.assertEqual('./bar', e.package_path)

    def test_validate_person(self):
        auth1 = Person('foo')
        auth1.email = 'foo@bar.com'
        auth1.validate()
        auth1.email = 'foo-bar@bar.com'
        auth1.validate()
        auth1.email = 'foo+bar@bar.com'
        auth1.validate()

        auth1.email = 'foo[at]bar.com'
        self.assertRaises(InvalidPackage, Person.validate, auth1)
        auth1.email = 'foo bar.com'
        self.assertRaises(InvalidPackage, Person.validate, auth1)
        auth1.email = 'foo<bar.com'
        self.assertRaises(InvalidPackage, Person.validate, auth1)

    def test_check_known_attributes(self):

        def create_node(tag, attrs):
            data = '<%s %s/>' % (tag, ' '.join(('%s="%s"' % p) for p in attrs.items()))
            return dom.parseString(data).firstChild

        try:
            create_node('tag', {'key': 'value'})
        except ExpatError as e:
            self.fail('create_node() raised %s "%s" unexpectedly!' % (type(e), str(e)))

        self.assertRaisesRegex(Exception, 'unbound prefix: line 1, column 0', create_node, 'tag', {'ns:key': 'value'})

        try:
            create_node('tag', {'ns:key': 'value', 'xmlns:ns': 'urn:ns'})
        except ExpatError as e:
            self.fail('create_node() raised %s "%s" unexpectedly!' % (type(e), str(e)))

        def check(attrs, known, res=[]):
            self.assertEqual(res, _check_known_attributes(create_node('tag', attrs), known))

        expected_err = ['The "tag" tag must not have the following attributes: attr2']

        check({}, [])
        check({}, ['attr'])
        check({'attr': 'value'}, ['attr'])
        check({'attr2': 'value'}, ['attr'], expected_err)

        check({'xmlns': 'urn:ns'}, ['attr'])
        check({'xmlns:ns': 'urn:ns'}, ['attr'])
        check({'xmlns:ns': 'urn:ns', 'ns:attr': 'value'}, ['attr'])
        check({'xmlns:ns': 'urn:ns', 'ns:attr': 'value', 'attr2': 'value'}, ['attr'], expected_err)

    def test_parse_package_valid(self):
        filename = os.path.join(test_data_dir, 'valid_package.xml')
        package = parse_package(filename)
        assert package.filename == filename
        assert not package.is_metapackage()
        assert package.name == 'valid_package'
        assert package.description == 'valid_package description'
        assert package.version == '0.1.0'
        assert package.licenses == ['BSD']
        assert [x.name for x in package.run_depends] == ['foo', 'bar', 'baz']

    def test_parse_package_invalid(self):
        filename = os.path.join(test_data_dir, 'invalid_package.xml')
        self.assertRaises(InvalidPackage, parse_package, filename)

    def test_parse_package_xhtml_description(self):
        filename = os.path.join(test_data_dir, 'xhtml_description.xml')
        expected_plaintext_description = None
        with open(os.path.join(test_data_dir, 'xhtml_description.txt'), 'r') as f:
            # Strip the trailing newline from the data file.
            expected_plaintext_description = f.read().rstrip('\n')
        package = parse_package(filename)
        assert package.description

        assert package.plaintext_description == expected_plaintext_description

    def test_parse_package_string(self):
        filename = os.path.join(test_data_dir, 'valid_package.xml')
        xml = _get_package_xml(filename)[0]

        assert isinstance(xml, str)
        parse_package_string(xml)

        if sys.version_info[0] == 2:
            xml = xml.decode('utf-8')
            assert not isinstance(xml, str)
        else:
            xml = xml.encode('utf-8')
            assert isinstance(xml, bytes)
        parse_package_string(xml)

        xml_string = """
<package>
  <name>valid_package</name>
  <version>0.1.0</version>
  <description>valid_package description</description>
  <maintainer email="user@todo.todo>Forgotten end quote</maintainer>
  <license>BSD</license>
</package>
"""
        self.assertRaises(InvalidPackage, parse_package_string, xml_string)

        xml_string = """
<package>
  <name>valid_package</name>
  <version>0.1.0</version>
  <description>Invalid < character in description</description>
  <maintainer email="user@todo.todo">user</maintainer>
  <license>BSD</license>
</package>
"""
        self.assertRaises(InvalidPackage, parse_package_string, xml_string)
        xml_string = """
<package>
  <name>valid_package</name>
  <version>0.1.0</version>
  <description>valid_package description</description>
  <maintainer email="user@todo.todo">user</maintainer>
  <license>BSD</license>
</package><extra>Unwanted junk</extra>
"""
        self.assertRaises(InvalidPackage, parse_package_string, xml_string)

    def test_has_ros_schema_reference_string(self):
        self.assertFalse(
            has_ros_schema_reference_string(
                """\
<?xml version="1.0"?>
<?xml-model href="http://some.url/to/a_wrong_schema.xsd" schematypens="http://www.w3.org/2001/XMLSchema"?>
<package/>
"""))
        self.assertFalse(
            has_ros_schema_reference_string(
                """\
<?xml version="1.0"?>
<package/>
"""))
        for format_version in (1, 2, 3):
            self.assertTrue(
                has_ros_schema_reference_string(
                    """\
<?xml version="1.0"?>
<?xml-model href="http://download.ros.org/schema/package_format%d.xsd" schematypens="http://www.w3.org/2001/XMLSchema"?>
<package/>
""" % format_version))



import datetime
import os
import shutil
import tempfile
import unittest

from catkin_pkg.package_version import _replace_setup_py_version
from catkin_pkg.package_version import _replace_version
from catkin_pkg.package_version import bump_version
from catkin_pkg.package_version import update_changelog_sections
from catkin_pkg.package_version import update_versions

try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

from .util import in_temporary_directory


class PackageVersionTest(unittest.TestCase):

    def test_bump_version(self):
        self.assertEqual('0.0.1', bump_version('0.0.0'))
        self.assertEqual('1.0.1', bump_version('1.0.0'))
        self.assertEqual('0.1.1', bump_version('0.1.0'))
        self.assertEqual('0.0.1', bump_version('0.0.0', 'patch'))
        self.assertEqual('1.0.1', bump_version('1.0.0', 'patch'))
        self.assertEqual('0.1.1', bump_version('0.1.0', 'patch'))
        self.assertEqual('1.0.0', bump_version('0.0.0', 'major'))
        self.assertEqual('1.0.0', bump_version('0.0.1', 'major'))
        self.assertEqual('1.0.0', bump_version('0.1.1', 'major'))
        self.assertEqual('0.1.0', bump_version('0.0.0', 'minor'))
        self.assertEqual('0.1.0', bump_version('0.0.1', 'minor'))
        self.assertEqual('1.1.0', bump_version('1.0.1', 'minor'))
        self.assertRaises(ValueError, bump_version, '0.0.asd')
        self.assertRaises(ValueError, bump_version, '0.0')
        self.assertRaises(ValueError, bump_version, '0')
        self.assertRaises(ValueError, bump_version, '0.0.-1')

    def test_replace_version(self):
        self.assertEqual('<package><version>0.1.1</version></package>',
                         _replace_version('<package><version>0.1.0</version></package>', '0.1.1'))
        self.assertEqual("<package><version abi='0.1.0'>0.1.1</version></package>",
                         _replace_version("<package><version abi='0.1.0'>0.1.0</version></package>", '0.1.1'))
        self.assertRaises(RuntimeError, _replace_version, '<package></package>', '0.1.1')
        self.assertRaises(RuntimeError, _replace_version, '<package><version>0.1.1</version><version>0.1.1</version></package>', '0.1.1')

    def test_replace_setup_py_version(self):
        self.assertEqual(
            'version="1.0.0",',
            _replace_setup_py_version('version="0.0.1",', '1.0.0'))
        self.assertEqual(
            "version='1.0.0',",
            _replace_setup_py_version("version='0.0.1',", '1.0.0'))
        self.assertRaises(
            RuntimeError,
            _replace_setup_py_version,
            '',
            '1.0.0')
        self.assertRaises(
            RuntimeError,
            _replace_setup_py_version,
            "version='0.1'",
            '1.0.0')
        self.assertRaises(
            RuntimeError,
            _replace_setup_py_version,
            'version=something,',
            '1.0.0')
        self.assertRaises(
            RuntimeError,
            _replace_setup_py_version,
            'version="0.0.1",version="0.0.1",',
            '1.0.0')

    def test_update_versions(self):
        try:
            root_dir = tempfile.mkdtemp()
            sub_dir = os.path.join(root_dir, 'sub')
            with open(os.path.join(root_dir, 'package.xml'), 'w') as fhand:
                fhand.write('<package><version>2.3.4</version></package>')
            os.makedirs(os.path.join(sub_dir))
            with open(os.path.join(sub_dir, 'package.xml'), 'w') as fhand:
                fhand.write('<package><version>1.5.4</version></package>')

            update_versions({root_dir: Mock(), sub_dir: Mock()}, '7.6.5')

            with open(os.path.join(root_dir, 'package.xml'), 'r') as fhand:
                contents = fhand.read()
                self.assertEqual('<package><version>7.6.5</version></package>', contents)
            with open(os.path.join(sub_dir, 'package.xml'), 'r') as fhand:
                contents = fhand.read()
                self.assertEqual('<package><version>7.6.5</version></package>', contents)
        finally:
            shutil.rmtree(root_dir)

    @in_temporary_directory
    def test_update_changelog_unicode(self, directory=None):
        """Test that updating the changelog does not throw an exception on unicode characters."""
        temp_file = os.path.join(directory, 'changelog')
        missing_changelogs_but_forthcoming = {}
        # Mock the Changelog object from catkin_pkg
        mock_changelog = Mock()
        # Create a changelog entry with a unicode char.
        mock_changelog.rst = ('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n'
                              'Changelog for package fake_pkg\n'
                              '^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n'
                              '\n'
                              'Forthcoming\n'
                              '-----------\n'
                              '* This is my changelog entry\n'
                              '* This is a line that has unicode' u'\xfc''\n'
                              '\n'
                              '0.0.9 (2017-01-30)\n'
                              '------------------\n'
                              '* This is old version.\n')

        # Create tuple with expected entires.
        missing_changelogs_but_forthcoming['fake_pkg'] = (temp_file, mock_changelog, 'Forthcoming')
        # Should not raise an exception
        update_changelog_sections(missing_changelogs_but_forthcoming, '1.0.0')

        # Generate dynamic lines, using present system date,
        # the length of the line of '-'s for the underline
        # and the utf-8 encoded data expected to be read back.
        ver_line = '1.0.0 (%s)' % datetime.date.today().isoformat()
        ver_line = ver_line.encode('utf-8')
        dash_line = '-' * len(ver_line)
        dash_line = dash_line.encode('utf-8')
        unicode_line = u'* This is a line that has unicode\xfc'.encode('utf-8')
        expected = [b'^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^',
                    b'Changelog for package fake_pkg',
                    b'^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^',
                    b'',
                    ver_line,
                    dash_line,
                    b'* This is my changelog entry',
                    unicode_line,
                    b'',
                    b'0.0.9 (2017-01-30)',
                    b'------------------',
                    b'* This is old version.']

        # Open the file written, and compare each line written to
        # the one read back.
        with open(temp_file, 'rb') as verify_file:
            content = verify_file.read().splitlines()
            for line_written, line_expected in zip(content, expected):
                self.assertEqual(line_written.strip(), line_expected)



from __future__ import print_function

import os
import shutil
import tempfile
import unittest

try:
    from catkin_pkg.workspaces import ensure_workspace_marker, get_spaces, order_paths,\
        CATKIN_WORKSPACE_MARKER_FILE
except ImportError as e:
    raise ImportError('Please adjust your PYTHONPATH before running this test: %s' % str(e))


class WorkspacesTest(unittest.TestCase):

    def test_ensure_workspace_marker(self):
        root_dir = tempfile.mkdtemp()
        try:
            ensure_workspace_marker(root_dir)
            self.assertTrue(os.path.exists(os.path.join(root_dir, CATKIN_WORKSPACE_MARKER_FILE)))
            # assert no exception on revisit
            ensure_workspace_marker(root_dir)
        finally:
            shutil.rmtree(root_dir)

    def test_get_spaces(self):
        self.assertEqual([], get_spaces([]))
        root_dir = tempfile.mkdtemp()
        try:
            self.assertEqual([], get_spaces([root_dir]))
            with open(os.path.join(root_dir, '.catkin'), 'a') as fhand:
                fhand.write('')
            self.assertEqual([root_dir], get_spaces([root_dir]))
        finally:
            shutil.rmtree(root_dir)

    def test_order_paths(self):
        self.assertEqual([], order_paths([], []))
        self.assertEqual(['bar', 'baz'], order_paths(['bar', 'baz'], ['foo']))
        self.assertEqual(['foo', 'bar'], order_paths(['bar', 'foo'], ['foo']))
        self.assertEqual(['baz', 'foo', 'bar'], order_paths(['bar', 'foo', 'baz'], ['baz', 'foo']))
        self.assertEqual(['foo' + os.sep + 'bim', 'bar'], order_paths(['bar', 'foo' + os.sep + 'bim'], ['foo']))

    def test_order_paths_with_symlink(self):
        root_dir = tempfile.mkdtemp()
        try:
            foo = os.path.join(root_dir, 'foo')
            foo_inc = os.path.join(foo, 'include')
            foo_ln = os.path.join(root_dir, 'foo_symlink')
            try:
                os.symlink(foo, foo_ln)
            except (AttributeError, OSError):
                self.skipTest('requires symlink availability')

            self.assertEqual([foo, 'bar'], order_paths(['bar', foo], [foo_ln]))
            self.assertEqual([foo_ln, 'bar'], order_paths(['bar', foo_ln], [foo]))
            self.assertEqual([foo_inc, 'bar'], order_paths(['bar', foo_inc], [foo_ln]))
        finally:
            shutil.rmtree(root_dir)



"""This script generates REP-0132 CHANGELOG.rst files for git or hg repositories."""

from __future__ import print_function

import argparse
import logging
import os
import sys

from catkin_pkg.changelog import CHANGELOG_FILENAME
from catkin_pkg.changelog_generator import generate_changelog_file, generate_changelogs, get_all_changes, get_forthcoming_changes, update_changelogs
from catkin_pkg.changelog_generator_vcs import get_vcs_client
from catkin_pkg.packages import find_packages

try:
    raw_input
except NameError:
    raw_input = input  # noqa: A001


def prompt_continue(msg, default):
    """Prompt the user for continuation."""
    if default:
        msg += ' [Y/n]?'
    else:
        msg += ' [y/N]?'

    while True:
        response = raw_input(msg)
        if not response:
            response = 'y' if default else 'n'
        else:
            response = response.lower()

        if response in ['y', 'n']:
            return response == 'y'

        print("Response '%s' was not recognized, please use one of the following options: y, Y, n, N" % response, file=sys.stderr)


def main(sysargs=None):
    parser = argparse.ArgumentParser(description='Generate a REP-0132 %s' % CHANGELOG_FILENAME)
    group_merge = parser.add_mutually_exclusive_group()
    parser.add_argument(
        '-a', '--all', action='store_true', default=False,
        help='Generate changelog for all versions instead of only the forthcoming one (only supported when no changelog file exists yet)')
    group_merge.add_argument(
        '--only-merges', action='store_true', default=False,
        help='Only add merge commits to the changelog')
    parser.add_argument(
        '--print-root', action='store_true', default=False,
        help='Output changelog content to the console as if there would be only one package in the root of the repository')
    parser.add_argument(
        '--skip-contributors', action='store_true', default=False,
        help='Skip adding the list of contributors to the changelog')
    group_merge.add_argument(
        '--skip-merges', action='store_true', default=False,
        help='Skip adding merge commits to the changelog')
    parser.add_argument(
        '-y', '--non-interactive', action='store_true', default=False,
        help="Run without user interaction, confirming all questions with 'yes'")
    args = parser.parse_args(sysargs)

    base_path = '.'
    logging.basicConfig(format='%(message)s', level=logging.DEBUG)

    vcs_client = get_vcs_client(base_path)

    if args.print_root:
        # printing status messages to stderr to allow piping the changelog to a file
        if args.all:
            print('Querying all tags and commit information...', file=sys.stderr)
            tag2log_entries = get_all_changes(vcs_client, skip_merges=args.skip_merges, only_merges=args.only_merges)
            print('Generating changelog output with all versions...', file=sys.stderr)
        else:
            print('Querying commit information since latest tag...', file=sys.stderr)
            tag2log_entries = get_forthcoming_changes(vcs_client, skip_merges=args.skip_merges, only_merges=args.only_merges)
            print('Generating changelog files with forthcoming version...', file=sys.stderr)
        print('', file=sys.stderr)
        data = generate_changelog_file('repository-level', tag2log_entries, vcs_client=vcs_client)
        print(data)
        return 0

    # find packages
    packages = find_packages(base_path)
    if not packages:
        raise RuntimeError('No packages found')
    print('Found packages: %s' % ', '.join(sorted(p.name for p in packages.values())))

    # check for missing changelogs
    missing_changelogs = []
    for pkg_path, package in packages.items():
        changelog_path = os.path.join(base_path, pkg_path, CHANGELOG_FILENAME)
        if not os.path.exists(changelog_path):
            missing_changelogs.append(package.name)

    if args.all and not missing_changelogs:
        raise RuntimeError('All packages already have a changelog. Either remove (some of) them before using --all or invoke the script without --all.')

    if args.all and len(missing_changelogs) != len(packages):
        ignored = set([p.name for p in packages.values()]) - set(missing_changelogs)
        print('The following packages already have a changelog file and will be ignored: %s' % ', '.join(sorted(ignored)), file=sys.stderr)

    # prompt to switch to --all
    if not args.all and missing_changelogs:
        print('Some of the packages have no changelog file: %s' % ', '.join(sorted(missing_changelogs)))
        print('You might consider to use --all to generate the changelogs for all versions (not only for the forthcoming version).')
        if not args.non_interactive and not prompt_continue('Continue without --all option', default=False):
            raise RuntimeError('Skipping generation, rerun the script with --all.')

    if args.all:
        print('Querying all tags and commit information...')
        tag2log_entries = get_all_changes(vcs_client, skip_merges=args.skip_merges, only_merges=args.only_merges)
        print('Generating changelog files with all versions...')
        generate_changelogs(base_path, packages, tag2log_entries, logger=logging, vcs_client=vcs_client, skip_contributors=args.skip_contributors)
    else:
        print('Querying commit information since latest tag...')
        tag2log_entries = get_forthcoming_changes(vcs_client, skip_merges=args.skip_merges, only_merges=args.only_merges)
        # separate packages with/without a changelog file
        packages_without = {pkg_path: package for pkg_path, package in packages.items() if package.name in missing_changelogs}
        if packages_without:
            print('Generating changelog files with forthcoming version...')
            generate_changelogs(base_path, packages_without, tag2log_entries, logger=logging, vcs_client=vcs_client, skip_contributors=args.skip_contributors)
        packages_with = {pkg_path: package for pkg_path, package in packages.items() if package.name not in missing_changelogs}
        if packages_with:
            print('Updating forthcoming section of changelog files...')
            update_changelogs(base_path, packages_with, tag2log_entries, logger=logging, vcs_client=vcs_client, skip_contributors=args.skip_contributors)
    print('Done.')
    print('Please review the extracted commit messages and consolidate the changelog entries before committing the files!')


def main_catching_runtime_error(*args, **kwargs):
    try:
        main(*args, **kwargs)
    except RuntimeError as e:
        print('ERROR: ' + str(e), file=sys.stderr)
        sys.exit(1)



import os

from catkin_pkg.package import InvalidPackage
from catkin_pkg.packages import find_package_paths
from catkin_pkg.packages import find_packages
from catkin_pkg.packages import find_packages_allowing_duplicates

from .util import in_temporary_directory


test_data_dir = os.path.join(os.path.dirname(__file__), 'data', 'ignored_packages')


def _create_pkg_in_dir(path, version='0.1.0'):
    path = os.path.abspath(path)
    os.makedirs(path)

    template = """\
<?xml version="1.0"?>
<package>
  <name>{0}</name>
  <version>{1}</version>
  <description>Package {0}</description>
  <license>BSD</license>

  <maintainer email="foo@bar.com">Foo Bar</maintainer>
</package>
""".format(path.split('/')[-1], version)

    with open(os.path.join(path, 'package.xml'), 'w+') as f:
        f.write(template)


@in_temporary_directory
def test_package_paths_with_hidden_directories():
    _create_pkg_in_dir('.test1')
    _create_pkg_in_dir('.test2')
    _create_pkg_in_dir('test3')  # not hidden
    _create_pkg_in_dir('.test4')

    res = find_package_paths('.')
    assert res == ['test3']


@in_temporary_directory
def test_find_packages_allowing_duplicates_with_no_packages():
    res = find_packages_allowing_duplicates('.')
    assert isinstance(res, dict)
    assert not res


@in_temporary_directory
def test_find_packages_invalid_version():
    version = ':{version}'
    path = 'src/foo'
    _create_pkg_in_dir(path, version)
    try:
        find_packages(path.split('/')[0])
        assert False, 'Must raise'
    except InvalidPackage as e:
        exception_message = str(e)
        assert version in exception_message
        assert path in exception_message


def test_find_no_ignored_packages():
    result = find_packages(test_data_dir)
    assert 'catkin_ignore' not in result
    assert 'custom_ignore' in result


def test_find_no_ignored_packages_with_custom_ignore():
    custom_result = find_packages(test_data_dir, ignore_markers={'CUSTOM_IGNORE'})
    assert 'custom_ignore' not in custom_result
    assert 'catkin_ignore' in custom_result



# Software License Agreement (BSD License)
#
# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Library for providing the relevant information from the package manifest for the Python setup.py file."""

from __future__ import print_function

import os
import sys

from .package import InvalidPackage, parse_package


def generate_distutils_setup(package_xml_path=os.path.curdir, **kwargs):
    """
    Extract the information relevant for distutils from the package manifest.

    The following keys will be set:

    The "name" and "version" are taken from the eponymous tags.

    A single maintainer will set the keys "maintainer" and
    "maintainer_email" while multiple maintainers are merged into the
    "maintainer" fields (including their emails). Authors are handled
    likewise.

    The first URL of type "website" (or without a type) is used for
    the "url" field.

    The "description" is taken from the eponymous tag if it does not
    exceed 200 characters and has no newlines. If it does "description"
    contains the truncated text while "long_description" contains the
    complete.

    All licenses are merged into the "license" field.

    :param kwargs: All keyword arguments are passed through. The above
        mentioned keys are verified to be identical if passed as a
        keyword argument

    :returns: return dict populated with parsed fields and passed
        keyword arguments
    :raises: :exc:`InvalidPackage`
    :raises: :exc:`IOError`
    """
    package = parse_package(package_xml_path)

    data = {}
    data['name'] = package.name
    data['version'] = package.version

    # either set one author with one email or join all in a single field
    if len(package.authors) == 1 and package.authors[0].email is not None:
        data['author'] = package.authors[0].name
        data['author_email'] = package.authors[0].email
    else:
        data['author'] = ', '.join([('%s <%s>' % (a.name, a.email) if a.email is not None else a.name) for a in package.authors])

    # either set one maintainer with one email or join all in a single field
    if len(package.maintainers) == 1:
        data['maintainer'] = package.maintainers[0].name
        data['maintainer_email'] = package.maintainers[0].email
    else:
        data['maintainer'] = ', '.join(['%s <%s>' % (m.name, m.email) for m in package.maintainers])

    # either set the first URL with the type 'website' or the first URL of any type
    websites = [url.url for url in package.urls if url.type == 'website']
    if websites:
        data['url'] = websites[0]
    elif package.urls:
        data['url'] = package.urls[0].url

    description = package.plaintext_description.splitlines()[0]
    if len(description) > 200:
        description = description[:197] + '...'

    data['description'] = description
    if description != package.plaintext_description:
        data['long_description'] = package.plaintext_description

    data['license'] = ', '.join(package.licenses)

    # pass keyword arguments and verify equality if generated and passed in
    for k, v in kwargs.items():
        if k in data:
            if v != data[k]:
                raise InvalidPackage('The keyword argument "%s" does not match the information from package.xml: "%s" != "%s"' % (k, v, data[k]), package_xml_path)
        else:
            data[k] = v

    return data


def get_global_bin_destination():
    return 'bin'


def get_global_etc_destination():
    return 'etc'


def get_global_include_destination():
    return 'include'


def get_global_lib_destination():
    return 'lib'


def get_global_libexec_destination():
    return 'lib'


def get_global_python_destination():
    dest = 'lib/python%u.%u/' % (sys.version_info[0], sys.version_info[1])
    if '--install-layout=deb' not in sys.argv[1:]:
        dest += 'site-packages'
    else:
        dest += 'dist-packages'
    return dest


def get_global_share_destination():
    return 'share'


def get_package_bin_destination(pkgname):
    return os.path.join(get_global_libexec_destination(), pkgname)


def get_package_etc_destination(pkgname):
    return os.path.join(get_global_etc_destination(), pkgname)


def get_package_include_destination(pkgname):
    return os.path.join(get_global_include_destination(), pkgname)


def get_package_lib_destination(_pkgname):
    return get_global_lib_destination()


def get_package_python_destination(pkgname):
    return os.path.join(get_global_python_destination(), pkgname)


def get_package_share_destination(pkgname):
    return os.path.join(get_global_share_destination(), pkgname)



"""This script finds a catkin packages."""

from __future__ import print_function

import argparse
import os
import sys

from catkin_pkg.packages import find_packages


def main(argv=sys.argv[1:]):
    parser = argparse.ArgumentParser(description='Find a catkin package')
    parser.add_argument('pkg', help='The name of the package')
    parser.add_argument('base_path', nargs='?', default=os.curdir, help='The base path to crawl for packages')

    args = parser.parse_args(argv)

    try:
        packages = find_packages(args.base_path)
        catkin_pkg = [path for path, p in packages.items() if p.name == args.pkg]
        if catkin_pkg:
            print(catkin_pkg[0])
        else:
            print("Could not find package '%s'." % args.pkg, file=sys.stderr)
            sys.exit(2)
    except RuntimeError as e:
        print('ERROR: ' + str(e), file=sys.stderr)
        sys.exit(1)



import os

from catkin_pkg.tool_detection import get_previous_tool_used_on_the_space
from catkin_pkg.tool_detection import mark_space_as_built_by

from .util import in_temporary_directory


@in_temporary_directory
def test_get_previous_tool_used_on_the_space():
    res = get_previous_tool_used_on_the_space('folder_that_does_not_exist')
    assert res is None, res
    os.makedirs('build')
    res = get_previous_tool_used_on_the_space('build')
    assert res is None, res
    mark_space_as_built_by('build', 'foo')
    res = get_previous_tool_used_on_the_space('build')
    assert res == 'foo', res



import unittest

from catkin_pkg.terminal_color import ansi, enable_ANSI_colors, fmt, sanitize

try:
    char = unichr
except NameError:
    char = chr


class TerminalColorTest(unittest.TestCase):

    def test_terminal_colors(self):
        # since other test might disable ansi colors
        # we need to ensure it is enabled
        enable_ANSI_colors()
        assert ansi('reset') != ''
        test = '@_This is underlined@|'
        rslt = '\033[4mThis is underlined\033[0m\033[0m'
        assert fmt(test) == rslt
        test = 'This has bad stuff @! @/ @_ @| OK!'
        test = sanitize(test)
        rslt = 'This has bad stuff @! @/ @_ @| OK!\033[0m'
        assert fmt(test) == rslt
        test = char(2018)
        test = sanitize(test)
        rslt = char(2018)
        assert test == rslt



# Software License Agreement (BSD License)
#
# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function

import copy
import os
import sys

from .packages import find_packages
from .workspaces import get_spaces


class _PackageDecorator(object):

    def __init__(self, package, path):
        self.package = package
        self.path = path
        self.is_metapackage = 'metapackage' in (e.tagname for e in self.package.exports)
        message_generators = [e.content for e in self.package.exports if e.tagname == 'message_generator']
        self.message_generator = message_generators[0] if message_generators else None
        # a set containing this package name, direct build depends
        # and recursive run_depends of these build_depends
        self.depends_for_topological_order = None
        # a set containing this package name and recursive run_depends
        self._recursive_run_depends_for_topological_order = None

    def __getattr__(self, name):
        if name.startswith('__'):
            raise AttributeError(name)
        return getattr(self.package, name)

    def calculate_depends_for_topological_order(self, packages):
        """
        Set self.depends_for_topological_order to the recursive dependencies required for topological order.

        It contains this package name, all direct build- and buildtool
        dependencies and their recursive runtime dependencies.
        The set only contains packages which are in the passed packages
        dictionary.

        :param packages: dict of name to ``_PackageDecorator``
        """
        self.depends_for_topological_order = set()
        all_depends = self.package.build_depends + self.package.buildtool_depends + self.package.test_depends
        names = [d.name for d in all_depends if d.evaluated_condition]

        # collect all group dependencies
        for group_depend in self.package.group_depends:
            if group_depend.evaluated_condition:
                assert group_depend.members is not None, \
                    'Group members need to be determined before'
                names += group_depend.members

        # skip external dependencies, meaning names that are not known packages
        for name in [n for n in names if n in packages.keys()]:
            if not self.is_metapackage and packages[name].is_metapackage:
                print('WARNING: package "%s" should not depend on metapackage "%s" but on its packages instead' % (self.name, name), file=sys.stderr)
            if name in self.depends_for_topological_order:
                # avoid function call to improve performance
                # check within the loop since the set changes every cycle
                continue
            packages[name]._add_recursive_run_depends(packages, self.depends_for_topological_order)

    def _add_recursive_run_depends(self, packages, depends_for_topological_order):
        """
        Modify depends_for_topological_order argument by adding run_depends of self recursively.

        Only packages which are in the passed packages are added and recursed into.

        :param packages: dict of name to ``_PackageDecorator``
        :param depends_for_topological_order: set to be extended
        """
        if self._recursive_run_depends_for_topological_order is None:
            self._recursive_run_depends_for_topological_order = set()
            self._recursive_run_depends_for_topological_order.add(self.package.name)
            package_names = packages.keys()
            names = [d.name for d in self.package.run_depends if d.evaluated_condition]

            for group_depend in self.package.group_depends:
                if group_depend.evaluated_condition:
                    assert group_depend.members is not None, \
                        'Group members need to be determined before'
                    names += group_depend.members

            for name in [n for n in names
                         if (n in package_names and
                             n not in self._recursive_run_depends_for_topological_order)]:
                packages[name]._add_recursive_run_depends(packages,
                                                          self._recursive_run_depends_for_topological_order)

        depends_for_topological_order.update(self._recursive_run_depends_for_topological_order)


def topological_order(root_dir, whitelisted=None, blacklisted=None, underlay_workspaces=None):
    """
    Crawls the filesystem to find packages and uses their dependencies to return a topologically order list.

    When a circular dependency is detected, the last item in the returned list
    is a tuple with None and a string giving a superset of the guilty packages.

    :param root_dir: The path to search in, ``str``
    :param whitelisted: A list of whitelisted package names, ``list``
    :param blacklisted: A list of blacklisted package names, ``list``
    :param underlay_workspaces: A list of underlay workspaces of packages which might provide dependencies in case of partial workspaces, ``list``
    :returns: A list of tuples containing the relative path and a ``Package`` object, ``list``
    """
    packages = find_packages(root_dir)

    # find packages in underlayed workspaces
    underlay_packages = {}
    if underlay_workspaces:
        for workspace in reversed(underlay_workspaces):
            # since underlay workspace might be a devel space
            # consider spaces stored in the .catkin file
            spaces = get_spaces([workspace])
            for space in spaces:
                for path, package in find_packages(space).items():
                    underlay_packages[package.name] = (path, package)

    return topological_order_packages(packages, whitelisted=whitelisted, blacklisted=blacklisted, underlay_packages=dict(underlay_packages.values()))


def topological_order_packages(packages, whitelisted=None, blacklisted=None, underlay_packages=None):
    """
    Topologically orders packages.

    evaluate_conditions() will be called for each package.

    If group dependencies haven't determined their members yet
    extract_group_members() will be called for each group dependency to do so.

    First returning packages which have message generators and then
    the rest based on direct build-/buildtool_depends and indirect
    recursive run_depends.

    When a circular dependency is detected, the last item in the returned list
    is a tuple with None and a string giving a superset of the guilty packages.

    :param packages: A dict mapping relative paths to ``Package`` objects ``dict``
    :param whitelisted: A list of whitelisted package names, ``list``
    :param blacklisted: A list of blacklisted package names, ``list``
    :param underlay_packages: A dict mapping relative paths to ``Package`` objects ``dict``
    :returns: A list of tuples containing the relative path and a ``Package`` object, ``list``
    """
    decorators_by_name = {}
    for path, package in packages.items():
        # skip non-whitelisted packages
        if whitelisted and package.name not in whitelisted:
            continue
        # skip blacklisted packages
        if blacklisted and package.name in blacklisted:
            continue
        if package.name in decorators_by_name:
            path_with_same_name = decorators_by_name[package.name].path
            raise RuntimeError('Two packages with the same name "%s" in the workspace:\n- %s\n- %s' % (package.name, path_with_same_name, path))
        decorators_by_name[package.name] = _PackageDecorator(package, path)

    underlay_decorators_by_name = {}
    if underlay_packages:
        for path, package in underlay_packages.items():
            # skip overlayed packages
            if package.name in decorators_by_name:
                continue
            underlay_decorators_by_name[package.name] = _PackageDecorator(package, path)
        decorators_by_name.update(underlay_decorators_by_name)

    # evaluate conditions and determine group membership
    pkgs = [d.package for d in decorators_by_name.values()]
    for pkg in pkgs:
        pkg.evaluate_conditions(os.environ)
    for pkg in pkgs:
        for group_depend in pkg.group_depends:
            if group_depend.evaluated_condition:
                group_depend.extract_group_members(pkgs)

    # calculate transitive dependencies
    for decorator in decorators_by_name.values():
        decorator.calculate_depends_for_topological_order(decorators_by_name)

    tuples = _sort_decorated_packages(decorators_by_name)
    # remove underlay packages from result
    return [(path, package) for path, package in tuples if path is None or package.name not in underlay_decorators_by_name]


def _reduce_cycle_set(packages_orig):
    """
    Remove iteratively some packages from a set that are definitely not part of any cycle.

    When there is a cycle in the package dependencies,
    _sort_decorated_packages only knows the set of packages containing
    the cycle.
    :param packages: A dict mapping package name to ``_PackageDecorator`` objects ``dict``
    :returns: A list of package names from the input which could not easily be detected as not being part of a cycle.
    """
    assert(packages_orig)
    packages = copy.copy(packages_orig)
    last_depended = None
    while len(packages) > 0:
        depended = set()
        for name, decorator in packages.items():
            if decorator.depends_for_topological_order:
                depended = depended.union(decorator.depends_for_topological_order)
        for name in list(packages.keys()):
            if name not in depended:
                del packages[name]
        if last_depended:
            if last_depended == depended:
                return packages.keys()
        last_depended = depended


def _sort_decorated_packages(packages_orig):
    """
    Sorts packages according to dependency ordering.

    First considering the message generators and their recursive dependencies
    and then the rest of the packages.
    When a circle is detected, a tuple with None and a string giving a
    superset of the guilty packages.

    :param packages: A dict mapping package name to ``_PackageDecorator`` objects ``dict``
    :returns: A List of tuples containing the relative path and a ``Package`` object ``list``
    """
    packages = copy.deepcopy(packages_orig)

    # mark all packages which are (recursively) dependent on by message generators
    dependency_names_to_follow = set([name for name, decorator in packages.items() if decorator.message_generator])
    not_marked_package_names = set(packages.keys()) - dependency_names_to_follow
    while dependency_names_to_follow:
        pkg_name = dependency_names_to_follow.pop()
        for name in packages[pkg_name].depends_for_topological_order:
            if name in not_marked_package_names:
                # mark package
                packages[name].message_generator = True
                not_marked_package_names.remove(name)
                # queue for recursion
                dependency_names_to_follow.add(name)

    ordered_packages = []
    while len(packages) > 0:
        # find all packages without build dependencies
        message_generators = []
        non_message_generators = []
        for name, decorator in packages.items():
            if not decorator.depends_for_topological_order:
                if decorator.message_generator:
                    message_generators.append(name)
                else:
                    non_message_generators.append(name)
        # first choose message generators
        if message_generators:
            names = message_generators
        elif non_message_generators:
            names = non_message_generators
        else:
            # in case of a circular dependency pass a string with
            # the names list of remaining package names, with path
            # None to indicate cycle
            ordered_packages.append([None, ', '.join(sorted(_reduce_cycle_set(packages)))])
            break

        # alphabetic order only for convenience
        names.sort()

        # add first candidates to ordered list
        # do not add all candidates since removing the depends from the first might affect the next candidates
        name = names[0]
        ordered_packages.append([packages[name].path, packages[name].package])
        # remove package from further processing
        del packages[name]
        for package in packages.values():
            if name in package.depends_for_topological_order:
                package.depends_for_topological_order.remove(name)
    return ordered_packages



# Software License Agreement (BSD License)
#
# Copyright (c) 2013, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Open Source Robotics Foundation, Inc. nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Generate/update ROS changelog files.

The Changelog format is described in REP-0132:

http://ros.org/reps/rep-0132.html
"""

import os
import re

from catkin_pkg.changelog import CHANGELOG_FILENAME
from catkin_pkg.changelog_generator_vcs import Tag

FORTHCOMING_LABEL = 'Forthcoming'


def get_all_changes(vcs_client, skip_merges=False, only_merges=False):
    tags = _get_version_tags(vcs_client)

    # query all log entries per tag range
    tag2log_entries = {}
    previous_tag = Tag(None)
    for tag in sorted_tags(tags):
        log_entries = vcs_client.get_log_entries(
            from_tag=previous_tag.name, to_tag=tag.name, skip_merges=skip_merges, only_merges=only_merges)
        tag2log_entries[previous_tag] = log_entries
        previous_tag = tag
    log_entries = vcs_client.get_log_entries(
        from_tag=previous_tag.name, to_tag=None, skip_merges=skip_merges, only_merges=only_merges)
    tag2log_entries[previous_tag] = log_entries
    return tag2log_entries


def get_forthcoming_changes(vcs_client, skip_merges=False, only_merges=False):
    tags = _get_version_tags(vcs_client)
    latest_tag_name = _get_latest_version_tag_name(vcs_client)

    # query log entries since latest tag only
    tag2log_entries = {}
    from_tag = Tag(None)
    to_tag = Tag(latest_tag_name)
    for tag in sorted_tags(tags):
        if to_tag.name is None:
            to_tag = tag
        # ignore non-forthcoming log entries but keep version to identify injection point of forthcoming
        tag2log_entries[tag] = None
    log_entries = vcs_client.get_log_entries(
        from_tag=from_tag.name, to_tag=to_tag.name, skip_merges=skip_merges, only_merges=only_merges)
    tag2log_entries[from_tag] = log_entries
    return tag2log_entries


def _get_version_tags(vcs_client):
    # get all tags in descending order
    tags = vcs_client.get_tags()
    version_tags = [t for t in tags if re.match(r'^v?\d+\.\d+.\d+$', t.name)]
    return version_tags


def _get_latest_version_tag_name(vcs_client):
    # get latest tag
    tag_name = vcs_client.get_latest_tag_name()
    if not re.match(r'^v?\d+\.\d+.\d+$', tag_name):
        raise RuntimeError(
            "The tag name '{}' doesn't match the version pattern v?x.y.z".format(tag_name))
    return tag_name


def generate_changelogs(base_path, packages, tag2log_entries, logger=None, vcs_client=None, skip_contributors=False):
    for pkg_path, package in packages.items():
        changelog_path = os.path.join(base_path, pkg_path, CHANGELOG_FILENAME)
        if os.path.exists(changelog_path):
            continue
        # generate package specific changelog file
        if logger:
            logger.debug("- creating '%s'" % os.path.join(pkg_path, CHANGELOG_FILENAME))
        pkg_tag2log_entries = filter_package_changes(tag2log_entries, pkg_path)
        data = generate_changelog_file(package.name, pkg_tag2log_entries, vcs_client=vcs_client, skip_contributors=skip_contributors)
        with open(changelog_path, 'wb') as f:
            f.write(data.encode('utf-8'))


def update_changelogs(base_path, packages, tag2log_entries, logger=None, vcs_client=None, skip_contributors=False):
    for pkg_path in packages.keys():
        # update package specific changelog file
        if logger:
            logger.debug("- updating '%s'" % os.path.join(pkg_path, CHANGELOG_FILENAME))
        pkg_tag2log_entries = filter_package_changes(tag2log_entries, pkg_path)
        changelog_path = os.path.join(base_path, pkg_path, CHANGELOG_FILENAME)
        with open(changelog_path, 'rb') as f:
            data = f.read().decode('utf-8')
        data = update_changelog_file(data, pkg_tag2log_entries, vcs_client=vcs_client, skip_contributors=skip_contributors)
        with open(changelog_path, 'wb') as f:
            f.write(data.encode('utf-8'))


def filter_package_changes(tag2log_entries, pkg_path):
    pkg_tag2log_entries = {}
    # collect all log entries relevant for this package
    for tag, log_entries in tag2log_entries.items():
        if log_entries is None:
            pkg_log_entries = None
        else:
            pkg_log_entries = []
            for log_entry in log_entries:
                if log_entry.affects_path(pkg_path):
                    pkg_log_entries.append(log_entry)
        pkg_tag2log_entries[tag] = pkg_log_entries
    return pkg_tag2log_entries


def generate_changelog_file(pkg_name, tag2log_entries, vcs_client=None, skip_contributors=False):
    blocks = []
    blocks.append(generate_package_headline(pkg_name))

    for tag in sorted_tags(tag2log_entries.keys()):
        log_entries = tag2log_entries[tag]
        if log_entries is not None:
            blocks.append(generate_version_block(version_from_tag(tag.name), tag.timestamp, log_entries, vcs_client=vcs_client, skip_contributors=skip_contributors))

    return '\n'.join(blocks)


def update_changelog_file(data, tag2log_entries, vcs_client=None, skip_contributors=False):
    tags = sorted_tags(tag2log_entries.keys())
    for i, tag in enumerate(tags):
        log_entries = tag2log_entries[tag]
        if log_entries is None:
            continue
        content = generate_version_content(log_entries, vcs_client=vcs_client, skip_contributors=skip_contributors)

        # check if version section exists
        match = get_version_section_match(data, version_from_tag(tag.name))
        if match:
            # prepend content to existing section
            data = prepend_version_content(data, version_from_tag(tag.name), content)
            assert data is not None
        else:
            # find injection point of earliest following version
            for next_tag in list(tags)[i:]:
                match = get_version_section_match(data, version_from_tag(next_tag.name))
                if match:
                    block = generate_version_block(version_from_tag(tag.name), tag.timestamp, log_entries, vcs_client=vcs_client, skip_contributors=skip_contributors)
                    data = data[:match.start()] + block + '\n' + data[match.start():]
                    break
            if not match:
                if tag.name is None:
                    raise RuntimeError('Could not find section "%s"' % next_tag.name)
                else:
                    raise RuntimeError('Could neither find section "%s" nor any other section' % tag.name)
        return data


def get_version_section_match(data, version):
    pattern = get_version_section_pattern(version)
    matches = re.finditer(pattern, data, flags=re.MULTILINE)
    matches = list(matches)
    if len(matches) > 1:
        raise RuntimeError('Found multiple matching sections')
    return matches[0] if matches else None


def get_version_section_pattern(version):
    valid_section_characters = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
    headline = get_version_headline(version, None)
    pattern = '^(' + re.escape(headline) + r'( \([0-9 \-:|+]+\))?)\r?\n([' + re.escape(valid_section_characters) + ']+)\r?\n?$'
    return pattern


def prepend_version_content(data, version, content):
    pattern = get_version_section_pattern(version)

    def replace_section(match):
        headline = match.group(1)
        section = match.group(3)
        data = content.rstrip()
        if data:
            data += '\n'
        return headline + '\n' + section + '\n' + data

    data, count = re.subn(pattern, replace_section, data, flags=re.MULTILINE)
    if count > 1:
        raise RuntimeError('Found multiple matching sections')
    return data if count == 1 else None


def version_from_tag(tag_name):
    if tag_name is None:
        return None
    if tag_name.startswith('v'):
        return tag_name[1:]
    return tag_name


def sorted_tags(tags):
    # first return the forthcoming tag
    for tag in tags:
        if not tag.name:
            yield tag
    # then return the tags in descending order
    name_and_tag = [(t.name, t) for t in tags if t.name]
    name_and_tag.sort(key=lambda x: [int(y) for y in version_from_tag(x[0]).split('.')])
    name_and_tag.reverse()
    for (_, tag) in name_and_tag:
        yield tag


def generate_package_headline(pkg_name):
    headline = 'Changelog for package %s' % pkg_name
    section_marker = '^' * len(headline)
    return '%s\n%s\n%s\n' % (section_marker, headline, section_marker)


def generate_version_block(version, timestamp, log_entries, vcs_client=None, skip_contributors=False):
    data = generate_version_headline(version, timestamp)
    data += generate_version_content(log_entries, vcs_client=vcs_client, skip_contributors=skip_contributors)
    return data


def generate_version_headline(version, timestamp):
    headline = get_version_headline(version, timestamp)
    return '%s\n%s\n' % (headline, '-' * len(headline))


def get_version_headline(version, timestamp):
    if not version:
        return FORTHCOMING_LABEL
    headline = version
    if timestamp:
        headline += ' (%s)' % timestamp
    return headline


def generate_version_content(log_entries, vcs_client=None, skip_contributors=False):
    data = ''
    all_authors = set()
    for entry in log_entries:
        msg = entry.msg
        lines = msg.splitlines()
        lines = [line.strip() for line in lines]
        lines = [line for line in lines if line and not line.startswith('Signed-off-by:')]
        lines = [escape_trailing_underscores(line) for line in lines]
        data += '* %s\n' % (replace_repository_references(lines[0], vcs_client=vcs_client) if lines else '')
        for line in lines[1:]:
            data += '  %s\n' % replace_repository_references(line, vcs_client=vcs_client)
        all_authors.add(entry.author)
    if all_authors and not skip_contributors:
        data += '* Contributors: %s\n' % ', '.join(sorted(all_authors))
    return data


def escape_trailing_underscores(line):
    if line.endswith('_'):
        line = line[:-1] + r'\_'
    # match words ending with an underscore which are not followed by another word
    # and insert a backslash before the underscore to escape it
    line = re.sub(r'(\w+)_([^\w])', '\\1\\_\\2', line)
    return line


def replace_repository_references(line, vcs_client=None):
    if vcs_client:
        line = vcs_client.replace_repository_references(line)
    return line



# Software License Agreement (BSD License)
#
# Copyright (c) 2013, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Open Source Robotics Foundation, Inc. nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function

import os
import re


def get_metapackage_cmake_template_path():
    """
    Return the location of the metapackage CMakeLists.txt CMake template.

    :returns: ``str`` location of the metapackage CMakeLists.txt CMake template
    """
    rel_path = os.path.join('templates', 'metapackage.cmake.in')
    return os.path.join(os.path.dirname(__file__), rel_path)


def configure_file(template_file, environment):
    """
    Evaluate a .in template file used in CMake with configure_file().

    :param template_file: path to the template, ``str``
    :param environment: dictionary of placeholders to substitute,
      ``dict``
    :returns: string with evaluates template
    :raises: KeyError for placeholders in the template which are not
      in the environment
    """  # noqa: D402
    with open(template_file, 'r') as f:
        template = f.read()
        return configure_string(template, environment)


def configure_string(template, environment):
    """
    Substitute variables enclosed by @ characters.

    :param template: the template, ``str``
    :param environment: dictionary of placeholders to substitute,
      ``dict``
    :returns: string with evaluates template
    :raises: KeyError for placeholders in the template which are not
      in the environment
    """
    def substitute(match):
        var = match.group(0)[1:-1]
        return environment[var]
    return re.sub('@[a-zA-Z0-9_]+@', substitute, template)



"""This script creates the skeletton of a catkin package."""

from __future__ import print_function

import argparse
import os
import sys

from catkin_pkg.package_templates import create_package_files, PackageTemplate


def main(argv=sys.argv[1:], parent_path=os.getcwd()):
    parser = argparse.ArgumentParser(
        description='Creates a new catkin package')
    parser.add_argument('name',
                        nargs=1,
                        help='The name for the package')
    parser.add_argument('--meta',
                        action='store_true',
                        help='Creates meta-package files')
    parser.add_argument('dependencies',
                        nargs='*',
                        help='Catkin package Dependencies')
    parser.add_argument('-s', '--sys-deps',
                        nargs='*',
                        help='System Dependencies')
    parser.add_argument('-b', '--boost-comps',
                        nargs='*',
                        help='Boost Components')
    parser.add_argument('-V', '--pkg_version',
                        action='store',
                        help='Initial Package version')
    parser.add_argument('-D', '--description',
                        action='store',
                        help='Description')
    parser.add_argument('-l', '--license',
                        action='append',
                        help='Name for License, (e.g. BSD, MIT, GPLv3...)')
    parser.add_argument('-a', '--author',
                        action='append',
                        help='A single author, may be used multiple times')
    parser.add_argument('-m', '--maintainer',
                        action='append',
                        help='A single maintainer, may be used multiple times')
    rosdistro_name = os.environ['ROS_DISTRO'] if 'ROS_DISTRO' in os.environ else None
    parser.add_argument('--rosdistro', required=rosdistro_name is None, default=rosdistro_name, help='The ROS distro (default: environment variable ROS_DISTRO if defined)')

    args = parser.parse_args(argv)

    try:
        package_name = args.name[0]
        target_path = os.path.join(parent_path, package_name)
        package_template = PackageTemplate._create_package_template(
            package_name=package_name,
            description=args.description,
            licenses=args.license or [],
            maintainer_names=args.maintainer,
            author_names=args.author,
            version=args.pkg_version,
            catkin_deps=args.dependencies,
            system_deps=args.sys_deps,
            boost_comps=args.boost_comps)
        create_package_files(target_path=target_path,
                             package_template=package_template,
                             rosdistro=args.rosdistro,
                             newfiles={},
                             meta=args.meta)
        print('Successfully created files in %s. Please adjust the values in package.xml.' % target_path)
    except ValueError as vae:
        parser.error(str(vae))



# Copyright 2017 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator

import pyparsing as pp

# operatorPrecedence renamed to infixNotation in 1.5.7
try:
    from pyparsing import infixNotation
except ImportError:
    from pyparsing import operatorPrecedence as infixNotation


def evaluate_condition(condition, context):
    if condition is None:
        return True
    expr = _get_condition_expression()
    try:
        parse_results = expr.parseString(condition, parseAll=True)
    except pp.ParseException as e:
        raise ValueError(
            "condition '%s' failed to parse: %s" % (condition, e))
    return parse_results[0](context)


_condition_expression = None


def _get_condition_expression():
    global _condition_expression
    if not _condition_expression:
        operator = pp.Regex('==|!=|>=|>|<=|<').setName('operator')
        operator.setParseAction(_Operator)

        identifier = pp.Word('$', pp.alphanums + '_', min=2).setName('identifier')
        identifier.setParseAction(_Identifier)

        value = pp.Word(pp.alphanums + '_-').setName('value')
        value.setParseAction(_Value)

        double_quoted_value = pp.QuotedString('"').setName(
            'double_quoted_value')
        double_quoted_value.setParseAction(_Value)
        single_quoted_value = pp.QuotedString("'").setName(
            'single_quoted_value')
        single_quoted_value.setParseAction(_Value)

        comparison_term = identifier | value | double_quoted_value | \
            single_quoted_value

        condition = pp.Group(comparison_term + operator + comparison_term).setName('condition')
        condition.setParseAction(_Condition)

        _condition_expression = infixNotation(
            condition, [
                ('and', 2, pp.opAssoc.LEFT, _And),
                ('or', 2, pp.opAssoc.LEFT, _Or),
            ])
    return _condition_expression


class _Operator:
    operators = {
        '==': operator.eq,
        '!=': operator.ne,
        '<=': operator.le,
        '<': operator.lt,
        '>=': operator.ge,
        '>': operator.gt,
    }

    def __init__(self, t):
        self.value = t[0]

    def __call__(self, arg1, arg2, context):
        assert self.value in self.operators
        return self.operators[self.value](arg1(context), arg2(context))

    def __str__(self):
        return self.value

    __repr__ = __str__


class _Identifier:

    def __init__(self, t):
        self.value = t[0]

    def __call__(self, context):
        return str(context.get(self.value[1:], ''))

    def __str__(self):
        return self.value

    __repr__ = __str__


class _Value:

    def __init__(self, t):
        self.value = t[0]

    def __call__(self, context):
        return self.value

    def __str__(self):
        return self.value

    __repr__ = __str__


class _Condition:

    def __init__(self, t):
        self.value = t[0]

    def __call__(self, context):
        return self.value[1](self.value[0], self.value[2], context)

    def __str__(self):
        return ' '.join(map(str, self.value))

    __repr__ = __str__


class _BinOp:

    def __init__(self, t):
        self.args = t[0][0::2]

    def __call__(self, context):
        return self.evalop(a(context) for a in self.args)

    def __str__(self):
        sep = ' %s  ' % self.reprsymbol
        return '(' + sep.join(map(str, self.args)) + ')'

    __repr__ = __str__


class _And(_BinOp):
    reprsymbol = 'and'
    evalop = all


class _Or(_BinOp):
    reprsymbol = 'or'
    evalop = any



#!/usr/bin/env python

import os

from setuptools import setup


kwargs = {
    'name': 'catkin_pkg',
    # same version as in:
    # - src/catkin_pkg/__init__.py
    # - stdeb.cfg
    'version': '0.5.2',
    'packages': ['catkin_pkg', 'catkin_pkg.cli'],
    'package_dir': {'': 'src'},
    'package_data': {'catkin_pkg': ['templates/*.in']},
    'entry_points': {
        'console_scripts': [
            'catkin_create_pkg = catkin_pkg.cli.create_pkg:main',
            'catkin_find_pkg = catkin_pkg.cli.find_pkg:main',
            'catkin_generate_changelog = catkin_pkg.cli.generate_changelog:main_catching_runtime_error',
            'catkin_package_version = catkin_pkg.cli.package_version:main',
            'catkin_prepare_release = catkin_pkg.cli.prepare_release:main',
            'catkin_tag_changelog = catkin_pkg.cli.tag_changelog:main',
            'catkin_test_changelog = catkin_pkg.cli.test_changelog:main',
        ]},
    'author': 'Dirk Thomas',
    'author_email': 'dthomas@osrfoundation.org',
    'url': 'http://wiki.ros.org/catkin_pkg',
    'keywords': ['catkin', 'ROS'],
    'classifiers': [
        'Programming Language :: Python',
        'License :: OSI Approved :: BSD License'
    ],
    'description': 'catkin package library',
    'long_description': 'Library for retrieving information about catkin packages.',
    'license': 'BSD',
    'install_requires': [
        'docutils',
        'python-dateutil',
        'pyparsing',
        'setuptools',
    ],
    'extras_require': {
        'test': [
            'flake8',
            'flake8-blind-except',
            'flake8-builtins',
            'flake8-class-newline',
            'flake8-comprehensions',
            'flake8-deprecated',
            'flake8-docstrings',
            'flake8-import-order',
            'flake8-quotes',
            "mock; python_version < '3.3'",
            'pytest',
        ]},
}
if 'SKIP_PYTHON_MODULES' in os.environ:
    kwargs['packages'] = []
    kwargs['package_dir'] = {}
    kwargs['package_data'] = {}
if 'SKIP_PYTHON_SCRIPTS' in os.environ:
    kwargs['name'] += '_modules'
    kwargs['entry_points'] = {}

setup(**kwargs)



from __future__ import print_function

import argparse
import os
import sys

from catkin_pkg.package_version import bump_version
from catkin_pkg.package_version import update_versions
from catkin_pkg.packages import find_packages, verify_equal_package_versions

# find the import relatively if available to work before installing catkin or overlaying installed version
if os.path.exists(os.path.join(os.path.dirname(__file__), '..', 'python', 'catkin', '__init__.py')):
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'python'))


def main():
    parser = argparse.ArgumentParser(description='Show or bump the version number in package.xml files.')
    parser.add_argument('path', nargs='?', default='.', help='The path to a parent folder which contains package.xml files (default: .)')
    parser.add_argument('--bump', choices=('major', 'minor', 'patch'), help='Which part of the version number to bump?')
    args = parser.parse_args()

    try:
        packages = find_packages(args.path)
        if not packages:
            print('No packages found', file=sys.stderr)
            sys.exit(1)
        version = verify_equal_package_versions(packages.values())

        # only print the version number
        if args.bump is None:
            print(version)

        else:
            # bump the version number
            new_version = bump_version(version, args.bump)
            update_versions(packages, new_version)
            print('%s -> %s' % (version, new_version))
    except Exception as e:  # noqa: B902
        sys.exit(str(e))



from __future__ import print_function

import argparse
import os
import re
import subprocess
import sys

from catkin_pkg import metapackage
from catkin_pkg.changelog import CHANGELOG_FILENAME, get_changelog_from_path
from catkin_pkg.package import InvalidPackage, PACKAGE_MANIFEST_FILENAME
from catkin_pkg.package_version import bump_version
from catkin_pkg.package_version import get_forthcoming_label, update_changelog_sections, update_versions
from catkin_pkg.packages import find_packages, verify_equal_package_versions
from catkin_pkg.terminal_color import disable_ANSI_colors, fmt
from catkin_pkg.workspace_vcs import get_repository_type, vcs_remotes

try:
    from shutil import which
except ImportError:
    # fallback for Python < 3.3
    def which(exe):
        for path in os.getenv('PATH').split(os.path.pathsep):
            file_path = os.path.join(path, exe)
            if os.path.isfile(file_path):
                return file_path
        return None

try:
    raw_input
except NameError:
    raw_input = input  # noqa: A001


def has_changes(base_path, path, vcs_type):
    cmd = [_find_executable(vcs_type), 'diff', path]
    try:
        output = subprocess.check_output(cmd, cwd=base_path)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(fmt("@{rf}Failed to check if '@{boldon}%s@{boldoff}' has modifications: %s" % (path, str(e))))
    return output.decode('utf-8').rstrip() != ''


def prompt_continue(msg, default):
    """Prompt the user for continuation."""
    if default:
        msg += fmt(' @{yf}[Y/n]@{reset}?')
    else:
        msg += fmt(' @{yf}[y/N]@{reset}?')

    while True:
        _flush_stdin()
        try:
            response = raw_input(msg)
        except EOFError:
            response = ''
        if not response:
            response = 'y' if default else 'n'
        else:
            response = response.lower()

        if response in ['y', 'n']:
            return response == 'y'

        print(
            fmt(
                "@{yf}Response '@{boldon}%s@{boldoff}' was not recognized, please use one of the following options: %s" %
                (response, ', '.join([('@{boldon}%s@{boldoff}' % x) for x in ['y', 'Y', 'n', 'N']]))
            ), file=sys.stderr)


def _flush_stdin():
    try:
        from termios import tcflush, TCIFLUSH
        tcflush(sys.stdin, TCIFLUSH)
    except ImportError:
        # fallback if not supported on some platforms
        pass


def get_git_branch(base_path):
    cmd_branch = [_find_executable('git'), 'rev-parse', '--abbrev-ref', 'HEAD']
    try:
        branch = subprocess.check_output(cmd_branch, cwd=base_path)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(fmt('@{rf}Could not determine git branch: %s' % str(e)))
    return branch.decode('utf-8').rstrip()


def get_git_remote(base_path):
    branch = get_git_branch(base_path)

    cmd_remote = [_find_executable('git'), 'config', '--get', 'branch.%s.remote' % branch]
    try:
        remote = subprocess.check_output(cmd_remote, cwd=base_path)
    except subprocess.CalledProcessError as e:
        msg = 'Could not determine git remote: %s' % str(e)
        msg += "\n\nMay be the branch '%s' is not tracking a remote branch?" % branch
        raise RuntimeError(fmt('@{rf}%s' % msg))
    return remote.decode('utf-8').rstrip()


def try_repo_push(base_path, vcs_type):
    if vcs_type in ['git']:
        print('Trying to push to remote repository (dry run)...')
        cmd = [_find_executable(vcs_type), 'push']
        if vcs_type == 'git':
            cmd.extend(['-n'] + [get_git_remote(base_path), get_git_branch(base_path)])
        try:
            subprocess.check_call(cmd, cwd=base_path)
        except (subprocess.CalledProcessError, RuntimeError) as e:
            raise RuntimeError(fmt('@{rf}Failed to dry push to repository: %s' % str(e)))


def check_clean_working_copy(base_path, vcs_type):
    if vcs_type in ['bzr', 'hg', 'svn']:
        cmd = [_find_executable(vcs_type), 'status']
    elif vcs_type in ['git']:
        cmd = [_find_executable(vcs_type), 'status', '-s', '-u']
    else:
        assert False, 'Unknown vcs type: %s' % vcs_type
    try:
        output = subprocess.check_output(cmd, cwd=base_path)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(fmt('@{rf}Failed to check working copy state: %s' % str(e)))
    output = output.decode('utf-8').rstrip()
    if output != '':
        print(output)
        return False
    return True


def commit_files(base_path, vcs_type, packages, packages_with_changelogs, message, dry_run=False):
    cmd = [_find_executable(vcs_type), 'commit', '-m', message]
    cmd += [os.path.join(p, PACKAGE_MANIFEST_FILENAME) for p in packages.keys()]
    cmd += [s for s in [os.path.join(p, 'setup.py') for p in packages.keys()] if os.path.exists(s)]
    cmd += [path for path, _, _ in packages_with_changelogs.values()]
    if not dry_run:
        try:
            subprocess.check_call(cmd, cwd=base_path)
        except subprocess.CalledProcessError as e:
            raise RuntimeError(fmt('@{rf}Failed to commit package.xml files: %s' % str(e)))
    return cmd


def tag_repository(base_path, vcs_type, tag_name, has_tag_prefix, dry_run=False):
    if vcs_type in ['bzr', 'git', 'hg']:
        cmd = [_find_executable(vcs_type), 'tag', tag_name]
    elif vcs_type == 'svn':
        svn_url = vcs_remotes(base_path, 'svn')[5:]
        if os.path.basename(svn_url) == 'trunk':
            # tag "trunk"
            base_url = os.path.dirname(svn_url)
        elif os.path.basename(os.path.dirname(svn_url)) == 'branches':
            # tag a direct subfolder of "branches"
            base_url = os.path.dirname(os.path.dirname(svn_url))
        elif svn_url.rfind('/trunk/') != -1:
            # tag any subfolder of trunk but require a tag prefix
            if not has_tag_prefix:
                raise RuntimeError(fmt('@{rf}When tagging a subfolder you must use --tag-prefix to make your tag name unique'))
            base_url = svn_url[:svn_url.rfind('/trunk/')]
        elif svn_url.rfind('/branches/') != -1:
            # tag any subfolder of trunk but require a tag prefix
            if not has_tag_prefix:
                raise RuntimeError(fmt('@{rf}When tagging a subfolder you must use --tag-prefix to make your tag name unique'))
            base_url = svn_url[:svn_url.rfind('/branches/')]
        else:
            raise RuntimeError(fmt("@{rf}Could not determine base URL of SVN repository '%s'" % svn_url))
        tag_url = '%s/tags/%s' % (base_url, tag_name)
        cmd = ['svn', 'cp', '-m', '"tagging %s"' % tag_name, svn_url, tag_url]
    else:
        assert False, 'Unknown vcs type: %s' % vcs_type
    if not dry_run:
        try:
            subprocess.check_call(cmd, cwd=base_path)
        except subprocess.CalledProcessError as e:
            raise RuntimeError(fmt('@{rf}Failed to tag repository: %s' % str(e)))
    return cmd


def push_changes(base_path, vcs_type, tag_name, dry_run=False):
    commands = []

    # push changes to the repository
    cmd = [_find_executable(vcs_type), 'push']
    if vcs_type == 'git':
        cmd.extend([get_git_remote(base_path), get_git_branch(base_path)])
    commands.append(cmd)
    if not dry_run:
        try:
            subprocess.check_call(cmd, cwd=base_path)
        except subprocess.CalledProcessError as e:
            raise RuntimeError(fmt('@{rf}Failed to push changes to the repository: %s\n\nYou need to manually push the changes/tag to the repository.' % str(e)))

    # push tags to the repository
    if vcs_type in ['git']:
        cmd = [_find_executable(vcs_type), 'push', get_git_remote(base_path), tag_name]
        commands.append(cmd)
        if not dry_run:
            try:
                subprocess.check_call(cmd, cwd=base_path)
            except subprocess.CalledProcessError as e:
                raise RuntimeError(fmt('@{rf}Failed to push tag to the repository: %s\n\nYou need to manually push the new tag to the repository.' % str(e)))

    return commands


def _find_executable(vcs_type):
    file_path = which(vcs_type)
    if file_path is None:
        raise RuntimeError(fmt('@{rf}Could not find vcs binary: %s' % vcs_type))
    return file_path


def main():
    try:
        _main()
    except RuntimeError as e:
        print(e, file=sys.stderr)
        sys.exit(1)


def _main():
    parser = argparse.ArgumentParser(
        description='Runs the commands to bump the version number, commit the modified %s files and create a tag in the repository.' % PACKAGE_MANIFEST_FILENAME)
    parser.add_argument('--bump', choices=('major', 'minor', 'patch'), default='patch', help='Which part of the version number to bump? (default: %(default)s)')
    parser.add_argument('--version', help='Specify a specific version to use')
    parser.add_argument('--no-color', action='store_true', default=False, help='Disables colored output')
    parser.add_argument('--no-push', action='store_true', default=False, help='Disables pushing to remote repository')
    parser.add_argument('-t', '--tag-prefix', default='', help='Add this prefix to the created release tag')
    parser.add_argument('-y', '--non-interactive', action='store_true', default=False, help="Run without user interaction, confirming all questions with 'yes'")
    args = parser.parse_args()

    if args.version and not re.match('^(0|[1-9][0-9]*)\.(0|[1-9][0-9]*)\.(0|[1-9][0-9]*)$', args.version):
        parser.error('The passed version must follow the conventions (positive integers x.y.z with no leading zeros)')

    if args.tag_prefix and ' ' in args.tag_prefix:
        parser.error('The tag prefix must not contain spaces')

    # force --no-color if stdout is non-interactive
    if not sys.stdout.isatty():
        args.no_color = True
    # disable colors if asked
    if args.no_color:
        disable_ANSI_colors()

    base_path = '.'

    print(fmt('@{gf}Prepare the source repository for a release.'))

    # determine repository type
    vcs_type = get_repository_type(base_path)
    if vcs_type is None:
        raise RuntimeError(fmt("@{rf}Could not determine repository type of @{boldon}'%s'@{boldoff}" % base_path))
    print(fmt('Repository type: @{boldon}%s@{boldoff}' % vcs_type))

    # find packages
    try:
        packages = find_packages(base_path)
    except InvalidPackage as e:
        raise RuntimeError(fmt("@{rf}Invalid package at path @{boldon}'%s'@{boldoff}:\n  %s" % (os.path.abspath(base_path), str(e))))
    if not packages:
        raise RuntimeError(fmt('@{rf}No packages found'))
    print('Found packages: %s' % ', '.join([fmt('@{bf}@{boldon}%s@{boldoff}@{reset}' % p.name) for p in packages.values()]))

    # complain about packages with unsupported build_type as they might require additional steps before being released
    # complain about packages with upper case character since they won't be releasable with bloom
    unsupported_pkg_names = []
    invalid_pkg_names = []
    valid_build_types = ['catkin', 'ament_cmake', 'ament_python']
    for package in packages.values():
        build_types = package.get_unconditional_build_types()
        if any(build_type not in valid_build_types for build_type in build_types):
            unsupported_pkg_names.append(package.name)
        if package.name != package.name.lower():
            invalid_pkg_names.append(package.name)
    if unsupported_pkg_names:
        print(
            fmt(
                "@{yf}Warning: the following package are not of build_type %s and may require manual steps to release': %s" %
                (str(valid_build_types), ', '.join([('@{boldon}%s@{boldoff}' % p) for p in sorted(unsupported_pkg_names)]))
            ), file=sys.stderr)
        if not args.non_interactive and not prompt_continue('Continue anyway', default=False):
            raise RuntimeError(fmt('@{rf}Aborted release, verify that unsupported packages are ready to be released or release manually.'))
    if invalid_pkg_names:
        print(
            fmt(
                "@{yf}Warning: the following package names contain upper case characters which violate both ROS and Debian naming conventions': %s" %
                ', '.join([('@{boldon}%s@{boldoff}' % p) for p in sorted(invalid_pkg_names)])
            ), file=sys.stderr)
        if not args.non_interactive and not prompt_continue('Continue anyway', default=False):
            raise RuntimeError(fmt('@{rf}Aborted release, fix the names of the packages.'))

    local_modifications = []
    for pkg_path, package in packages.items():
        # verify that the package.xml files don't have modifications pending
        package_xml_path = os.path.join(pkg_path, PACKAGE_MANIFEST_FILENAME)
        if has_changes(base_path, package_xml_path, vcs_type):
            local_modifications.append(package_xml_path)
        # verify that metapackages are valid
        if package.is_metapackage():
            try:
                metapackage.validate_metapackage(pkg_path, package)
            except metapackage.InvalidMetapackage as e:
                raise RuntimeError(fmt(
                    "@{rf}Invalid metapackage at path '@{boldon}%s@{boldoff}':\n  %s\n\nSee requirements for metapackages: %s" %
                    (os.path.abspath(pkg_path), str(e), metapackage.DEFINITION_URL)))
        # verify that the setup.py files don't have modifications pending
        setup_py_path = os.path.join(pkg_path, 'setup.py')
        if os.path.exists(setup_py_path) and has_changes(base_path, setup_py_path, vcs_type):
            local_modifications.append(setup_py_path)

    # fetch current version and verify that all packages have same version number
    old_version = verify_equal_package_versions(packages.values())
    if args.version:
        new_version = args.version
    else:
        new_version = bump_version(old_version, args.bump)
    tag_name = args.tag_prefix + new_version

    if (
        not args.non_interactive and
        not prompt_continue(
            fmt(
                "Prepare release of version '@{bf}@{boldon}%s@{boldoff}@{reset}'%s" %
                (new_version, " (tagged as '@{bf}@{boldon}%s@{boldoff}@{reset}')" % tag_name if args.tag_prefix else '')
            ), default=True)
    ):
        raise RuntimeError(fmt("@{rf}Aborted release, use option '--bump' to release a different version and/or '--tag-prefix' to add a prefix to the tag name."))

    # check for changelog entries
    missing_changelogs = []
    missing_changelogs_but_forthcoming = {}
    for pkg_path, package in packages.items():
        changelog_path = os.path.join(pkg_path, CHANGELOG_FILENAME)
        if not os.path.exists(changelog_path):
            missing_changelogs.append(package.name)
            continue
        # verify that the changelog files don't have modifications pending
        if has_changes(base_path, changelog_path, vcs_type):
            local_modifications.append(changelog_path)
        changelog = get_changelog_from_path(changelog_path, package.name)
        try:
            changelog.get_content_of_version(new_version)
        except KeyError:
            # check that forthcoming section exists
            forthcoming_label = get_forthcoming_label(changelog.rst)
            if forthcoming_label:
                missing_changelogs_but_forthcoming[package.name] = (changelog_path, changelog, forthcoming_label)
            else:
                missing_changelogs.append(package.name)

    if local_modifications:
        raise RuntimeError(fmt('@{rf}The following files have modifications, please commit/revert them before:' + ''.join([('\n- @{boldon}%s@{boldoff}' % path) for path in local_modifications])))

    if missing_changelogs:
        print(
            fmt(
                "@{yf}Warning: the following packages do not have a changelog file or entry for version '@{boldon}%s@{boldoff}': %s" %
                (new_version, ', '.join([('@{boldon}%s@{boldoff}' % p) for p in sorted(missing_changelogs)]))
            ), file=sys.stderr)
        if not args.non_interactive and not prompt_continue('Continue without changelogs', default=False):
            raise RuntimeError(fmt("@{rf}Aborted release, populate the changelog with '@{boldon}catkin_generate_changelog@{boldoff}' and review / clean up the content."))

    # verify that repository is pushable (if the vcs supports dry run of push)
    if not args.no_push:
        try_repo_push(base_path, vcs_type)

    # check for staged changes and modified and untracked files
    print(fmt('@{gf}Checking if working copy is clean (no staged changes, no modified files, no untracked files)...'))
    is_clean = check_clean_working_copy(base_path, vcs_type)
    if not is_clean:
        print(fmt('@{yf}Warning: the working copy contains other changes. Consider reverting/committing/stashing them before preparing a release.'), file=sys.stderr)
        if not args.non_interactive and not prompt_continue('Continue anyway', default=False):
            raise RuntimeError(fmt('@{rf}Aborted release, clean the working copy before trying again.'))

    # for svn verify that we know how to tag that repository
    if vcs_type in ['svn']:
        tag_svn_cmd = tag_repository(base_path, vcs_type, tag_name, args.tag_prefix != '', dry_run=True)

    # tag forthcoming changelog sections
    update_changelog_sections(missing_changelogs_but_forthcoming, new_version)
    print(fmt(
        "@{gf}Rename the forthcoming section@{reset} of the following packages to version '@{bf}@{boldon}%s@{boldoff}@{reset}': %s" %
        (new_version, ', '.join([('@{boldon}%s@{boldoff}' % p) for p in sorted(missing_changelogs_but_forthcoming.keys())]))))

    # bump version number
    update_versions(packages, new_version)
    print(fmt("@{gf}Bump version@{reset} of all packages from '@{bf}%s@{reset}' to '@{bf}@{boldon}%s@{boldoff}@{reset}'" % (old_version, new_version)))

    pushed = None
    if vcs_type in ['svn']:
        # for svn everything affects the remote repository immediately
        commands = []
        commands.append(commit_files(base_path, vcs_type, packages, missing_changelogs_but_forthcoming, tag_name, dry_run=True))
        commands.append(tag_svn_cmd)
        if not args.no_push:
            print(fmt('@{gf}The following commands will be executed to commit the changes and tag the new version:'))
        else:
            print(fmt('@{gf}You can use the following commands to manually commit the changes and tag the new version:'))
        for cmd in commands:
            print(fmt('  @{bf}@{boldon}%s@{boldoff}' % ' '.join(cmd)))

        if not args.no_push:
            if not args.non_interactive:
                # confirm before modifying repository
                if not prompt_continue('Execute commands which will modify the repository', default=True):
                    pushed = False
            if pushed is None:
                commit_files(base_path, vcs_type, packages, missing_changelogs_but_forthcoming, tag_name)
                tag_repository(base_path, vcs_type, tag_name, args.tag_prefix != '')
                pushed = True

    else:
        # for other vcs types the changes are first done locally
        print(fmt('@{gf}Committing the package.xml files...'))
        commit_files(base_path, vcs_type, packages, missing_changelogs_but_forthcoming, tag_name)

        print(fmt("@{gf}Creating tag '@{boldon}%s@{boldoff}'..." % (tag_name)))
        tag_repository(base_path, vcs_type, tag_name, args.tag_prefix != '')

        try:
            commands = push_changes(base_path, vcs_type, tag_name, dry_run=True)
        except RuntimeError:
            print(fmt('@{yf}Warning: could not determine commands to push the changes and tag to the remote repository. Do you have a remote configured for the current branch?'))
        else:
            if not args.no_push:
                print(fmt('@{gf}The following commands will be executed to push the changes and tag to the remote repository:'))
            else:
                print(fmt('@{gf}You can use the following commands to manually push the changes to the remote repository:'))
            for cmd in commands:
                print(fmt('  @{bf}@{boldon}%s@{boldoff}' % ' '.join(cmd)))

            if not args.no_push:
                if not args.non_interactive:
                    # confirm commands to push to remote repository
                    if not prompt_continue('Execute commands to push the local commits and tags to the remote repository', default=True):
                        pushed = False
                if pushed is None:
                    push_changes(base_path, vcs_type, tag_name)
                    pushed = True

    if pushed:
        print(fmt("@{gf}The source repository has been released successfully. The next step will be '@{boldon}bloom-release@{boldoff}'."))
    else:
        msg = 'The release of the source repository has been prepared successfully but the changes have not been pushed yet. ' \
            "After pushing the changes manually the next step will be '@{boldon}bloom-release@{boldoff}'."
        if args.no_push or pushed is False:
            print(fmt('@{yf}%s' % msg))
        else:
            raise RuntimeError(fmt('@{rf}%s' % msg))
